//--------------------------------------------------------------------------
// REFFILE LIBRARY
//
// Package: reffile
//
// Description: A collection of routines for working with reference files
//              (FITS image or table) and
//              ASCII "loadfiles" in a uniform manner.
//
//              A vocabulary is adopted within these routines to describe 
//              structure and content across these file types.  
//              These definitions are used:
//
//                reffile - a FITS file or ASCII "loadfile"
//
//                row - FITS file extension number, FITS table row number or
//                      ASCII loadfile line number
//
//                header - FITS primary header or 
//                           ASCII loadfile "first keyword set"
//
//                body - FITS extension (image or table) or
//                           ASCII loadfile "Nth keyword set"
//
//                extension - can mean filename extension (like .fits or .lod)
//                              or FITS extension
//
// These routines are externally callable:
//
// access_reffile - determines if a file is accessible by trying to open it
// close_reffile - closes a file/object and frees its resources
// get_body_reffile - obtains a keyword value or table column row value from
//                      a file "body"
// get_hdr_reffile - obtains a keyword value from a file "header"
// nrow_reffile - returns the number of "groups" (image extensions) or rows in 
//                  an image, table, or loadfile
// open_reffile_hdr - try to open the primaryHDU or first group of a file
// open_reffile_body - try to open the 1st extension or second group of a file
// open_reffile - try to open an image, table, or loadfile for access
// put_body_reffile - add a table column value to a row within a "body"
// put_hdr_reffile - add a keyword value to a "header"
// put_hist_reffile - add a history comment to a reffile
// rename_reffile - renames a reference file
//
// These routines are only used within this package:
//
// count_reffile - determines number of "groups" in an image file
// reffile_name_has_ext - determine if name has extension (e.g. [n]) appended
// image_reffile - read image header keyword from the currently open extension
// reclose_reffile - close the current image/table/loadfile obj 
// reopen_reffile - attempt access the "row" requested, which may mean closing
//                    and reopening the file
// root_reffile - extract root (filename minus extension) from filename
//
//--------------------------------------------------------------------------
#include <stdexcept>
#include <fitsio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "system.h"
#include "c_iraf.h"
#include "xtables.h"
#include "ximio.h"
#include "c_iraf_priv.h"
#include "config.h"
#include "error.h"
#include "cdbslib.h"
#include "loadfile.h"
#include "cdbserror.h"
#include "AssertEx.h"
#include "reffile.h"

#define  NBITS_BYTE		8

#define  EXTRA_KEYS		"datamin|datamax|bitpix|datatype|simple|naxis1|naxis2"

#define  ORDINARY		0
#define  DATAMIN		1
#define  DATAMAX		2
#define  BITPIX			3
#define  DATATYPE		4
#define  SIMPLE			5
#define  NAXIS1			6
#define  NAXIS2			7

#define  MODELIST		"r|r+|w|a"

// internal functions
static int		count_reffile (reffile *);
static int		reffile_name_has_ext (char *);
static int		fits_reffile (char *);
static int		image_reffile (IRAFPointer, char *, char *, int);
static void 		reclose_reffile (reffile *);
static int 		reopen_reffile (reffile *, int);
static void		root_reffile (char *, char *, int);

//------------------------------------------------------------------------
// ACCESS_REFFILE -- Check access status of a reference file
//
// i: fname	file name
// i: type	file type (image, table, or load)
// i: mode	file access mode (see MODELIST)
//
//------------------------------------------------------------------------
int access_reffile (char *fname, char *ftype, char *mode)
{
    // the only reliable way found to check access 
    // to a file is to try to open it with the provided mode
    //
    reffile *ref = open_reffile (fname, ftype, mode);

    int status;
    if (ref == NULL) {
        status = NO;
    } else {
        close_reffile (ref);
	status = YES;
    }
    return (status);
}

//------------------------------------------------------------------------
// CLOSE_REFFILE -- Close the reference file and free the resources
//
// i: ref	reference file object
//
//------------------------------------------------------------------------
void close_reffile (reffile *ref) 
{
    reclose_reffile (ref);
    free (ref);
}

//------------------------------------------------------------------------
// GET_BODY_REFFILE -- Read a value from the "body" of a reference file
//
// u: ref	reference file object
// i: name	column name
// i: row	row number (zero indexed)
// o: value	keyword value
// i: maxch	max length of value
//
//------------------------------------------------------------------------
int get_body_reffile (reffile *ref, char *name, int row, 
		      char *value, int maxch)
{
    char	*str, *uname;
    int		status;
    IRAFPointer	cptr;

    // open reffile body (loadfile or FITS table/image extension)
    //   (WHY adding a +1 to the requested row?)
    //
    status = reopen_reffile (ref, row+1);
    if (status == ERR) {
        fprintf(stderr,"get_body_reffile: reopen_reffile failed\n");
        return (status);
    }

    // Read the value from the table body 
    switch (ref->type) {
    case IMAGE:
	status = image_reffile ((IRAFPointer) ref->ptr, name, value, maxch);
	break;
    case TABLE:
	c_tbcfnd1 ((IRAFPointer) ref->ptr, name, &cptr);
	if (cptr == 0) {
	    value[0] = EOS;
	    status = ERR;

	} else {
	    c_tbegtt ((IRAFPointer) ref->ptr, cptr, row+1, value, maxch);
	    if (c_iraferr () != 0) {
		value[0] = EOS;
		status = ERR;
	    }
	}
	break;
    case LOADFILE:
	uname = strclone (name);
	strucase (uname);

	str = get_tab_cell ((table *) ref->ptr, uname, row);
	if (str != NULL) {
	    safecopy (value, str, maxch);
	} else {
	    value[0] = EOS;
	    status = ERR;
	}

	free (uname);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    return (status);
}

//------------------------------------------------------------------------
// GET_HDR_REFFILE -- Read a header keyword from a reference file
//
// u: ref	reference file object
// i: name	keyword name
// o: value	keyword value
// i: maxch	max length of value
//
// 09/13/02  MSwam  add retry for 1st image extension
///
//------------------------------------------------------------------------
int get_hdr_reffile (reffile *ref, char *name, char *value, int maxch)
{
    char	*str, *uname;
    int		status;

    // Switch to reading primaryHDU
    status = reopen_reffile (ref, 0);
    if (status == ERR)
        return (status);

    // Read header keyword
    switch (ref->type) {
    case IMAGE:
	status = image_reffile ((IRAFPointer) ref->ptr, name, value, maxch);
        if (status == ERR) {
           // not in primaryHDU, retry using 1st image extension
           status = reopen_reffile (ref, 1);
           if (status == ERR) {
               fprintf(stderr,"ERROR-get_hdr_reffile: failed to open 1st "
                              "extension\n");
           }
           else {
	       status = image_reffile ((IRAFPointer) ref->ptr, name, 
                                                     value, maxch);
           }
        }
	break;
    case TABLE:
	c_tbhgtt ((IRAFPointer) ref->ptr, name, value, maxch);
	if (c_iraferr () != 0) {
	    value[0] = EOS;
	    status = ERR;
	}
	break;
    case LOADFILE:
	uname = strclone (name);
	strucase (uname);

	str = get_tab_key ((table *) ref->ptr, uname);
	if (str != NULL) {
	    safecopy (value, str, maxch);
	} else {
	    value[0] = EOS;
	    status = ERR;
	}

	free (uname);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    return (status);
}

//------------------------------------------------------------------------
// NROW_REFFILE -- Number of rows in a reference file
//
// i: ref	reference file object
//
//------------------------------------------------------------------------
int nrow_reffile (reffile *ref)
{
    int		nrow;
  
    Assert(ref != NULL, "nrow_reffile: BAD ref object");

    switch (ref->type) {
    case IMAGE:
	nrow = c_imgeti ((IRAFPointer) ref->ptr, "GCOUNT");
	if (c_iraferr () != 0)
	    nrow = count_reffile (ref);
	break;

    case TABLE:
        // read 1st table extension
        if (reopen_reffile (ref, 1) == ERR) {
	    cdbs_error (NOTOPEN, "%s", name_reffile (ref));

	} else {
	    nrow = c_tbpsta ((IRAFPointer) ref->ptr, TBL_NROWS);
	    if (c_iraferr () != 0)
	        cdbs_error (BADNROW, "%s", name_reffile (ref));
	}
	break;

    case LOADFILE:
	nrow = nrow_table ((table *) ref->ptr);
	break;
    }

    return (nrow);
}

//----------------------------------------------------------------------
// OPEN_REFFILE_HDR -- Open a reference file header
// OPEN_REFFILE_BODY -- Open a reference file body
//
// i: fname	reference file name
// i: type	reference file type (image, table, or load)
// i: mode      file open mode (e.g. read, write, etc. see MODELIST)
//
//----------------------------------------------------------------------
reffile *open_reffile_hdr (char *fname, char *type, char *mode)
{
    // If the file type = table, then in order for this routine to work we
    // need to FORCE the file open attempt as an IMAGE, since
    // we will be trying to read the primaryHDU for keyword values, and
    // opening the primaryHDU of a table as a table FAILS.
    //
    // If the file type != table (=image or load) then just open it as the
    //   requested type.
    //
    char phony_filetype[20];
    strcpy(phony_filetype, type);
    if (strcmp(type,"table")==0) {
       strcpy(phony_filetype, "image");
    }
    return (open_reffile (fname, phony_filetype, mode));
}

reffile *open_reffile_body (char *fname, char *type, char *mode)
{
    // slap a [1] on the name of a FITS file table to ensure that the 1st
    //   extension gets opened, rather than the default [0] primaryHDU
    //
    int flen = strlen(fname);
    char *adjusted_fname = (char *)malloc(flen+10);
    strcpy(adjusted_fname, fname);
    if (strcmp(type,"table")==0) {
       strcat(adjusted_fname,"[1]");
    }
    reffile *return_val = open_reffile (adjusted_fname, type, mode);
    free(adjusted_fname);
    return return_val;
}

//----------------------------------------------------------------------
// OPEN_REFFILE -- Open a reference file of the specified type
//
// i: fname	reference file name
// i: type	reference file type (image, table, or load)
// i: mode      file open mode (e.g. read, write, etc. see MODELIST)
//
//----------------------------------------------------------------------
reffile *open_reffile (char *fname, char *type, char *mode)
{
    // Allocate and intialize the object 
    reffile *ref = (reffile *) malloc (sizeof (reffile));
    Assert(ref != NULL, "open_reffile: BAD ref object");
    root_reffile (ref->name, fname, SZ_FNAME);
    ref->type = strabbrev (type, FILETYPE_NAMES);
    ref->mode = strmatch (mode, MODELIST);
    ref->is_fits = is_fitsname (ref->name);
    ref->smudge = NO;  // used to mark file content as changed(=YES)
    ref->ptr = NULL;

    // if no fits extension was specified in the filename, 
    //    assume primaryHDU or top row is intended
    //
    if ((ref->row = reffile_name_has_ext (fname)) == ERR)
        ref->row = 0;

    // Call reopen to actually open the file and requested extension/table row
    if (reopen_reffile (ref, ref->row) == ERR) {
        free (ref);
	ref = NULL;
    }

    // Return the object
    return (ref);
}

//------------------------------------------------------------------------
// PUT_BODY_REFFILE -- Write a value to the reference file body 
//
// u: ref	reference file object
// i: name	column name
// i: row	row number (zero indexed)
// i: value	cell value
///
//------------------------------------------------------------------------

void put_body_reffile (reffile *ref, char *name, int row, char *value)
{
    IRAFPointer	cptr;

    // Switch to accessing body 
    if (reopen_reffile (ref, row+1) == ERR)
        cdbs_error (NOWRITE, ref->name);

    switch (ref->type) {
    case IMAGE:
	cdbs_error (NOWRITE, "%s.%s", ref->name, name);
	break;
    case TABLE:
	c_tbcfnd1 ((IRAFPointer) ref->ptr, name, &cptr);
	if (cptr == 0) {
	    cdbs_error (NOWRITE, "%s.%s", ref->name, name);
	} else {
	    c_tbeptt ((IRAFPointer) ref->ptr, cptr, row+1, value);
	}
	break;
    case LOADFILE:
	put_tab_cell ((table *) ref->ptr, name, row, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->row = row;
    ref->smudge = YES;
}

//------------------------------------------------------------------------
// PUT_HDR_REFFILE -- Write a header keyword to a reference file
//
// u: ref	reference file object
// i: name	keyword name
// i: value	keyword value
//
//------------------------------------------------------------------------
void put_hdr_reffile (reffile *ref, char *name, char *value)
{
    // Switch to accessing primaryHDU
    if (reopen_reffile (ref, 0) == ERR)
        cdbs_error (NOWRITE, ref->name);

    // Write keyword value 
    switch (ref->type) {
    case IMAGE:
	c_imastr ((IRAFPointer) ref->ptr, name, value);
	break;
    case TABLE:
	c_tbhadt ((IRAFPointer) ref->ptr, name, value);
	break;
    case LOADFILE:
	put_tab_key ((table *) ref->ptr, name, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }
    ref->smudge = YES;
}

//------------------------------------------------------------------------
// PUT_HIST_REFFILE -- Write a history keyword to a reference file
//
// u: ref	reference file object
// i: name	keyword name
// i: value	keyword value
//
//------------------------------------------------------------------------
void put_hist_reffile (reffile *ref, char *name, char *value)
{
    // Switch to accessing primaryHDU
    if (reopen_reffile (ref, 0) == ERR)
        cdbs_error (NOWRITE, ref->name);

    // Write history keyword 
    switch (ref->type) {
    case IMAGE:
        c_imputh ((IRAFPointer) ref->ptr, name, value);
	break;
    case TABLE:
	c_tbhadt ((IRAFPointer) ref->ptr, name, value);
	break;
    case LOADFILE:
	put_tab_key ((table *) ref->ptr, name, value);
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->smudge = YES;
}

//------------------------------------------------------------------------
// RENAME_REFFILE -- Change the name of a reference file
//
// i: oldname	current file name
// i: newname	new file name
// i: type	file type
//
//------------------------------------------------------------------------
int rename_reffile (char *oldname, char *newname, char *ftype)
{
    int		type, status;

    status = OK;
    type = strabbrev (ftype, FILETYPE_NAMES);

    switch (type) {
    case IMAGE:
        // FOR NOW, assume that no special IMAGE file renaming is needed
        //   and do a straight file rename, as is done for the other types...
	// c_imrename (oldname, newname);
	// if (c_iraferr() != 0)
	//     status = ERR;
	// break;
    case TABLE:
    case LOADFILE:
	if (rename (oldname, newname) != 0)
	    status = ERR;
	break;
    default:
	cdbs_error (FTYPE, "%d", type);
    }

    return (status);
}

//------------------------------------------------------------------------
//
// Procedures below this line are not externally callable
//
//----------------------------------------------------------------------

//------------------------------------------------------------------------
// COUNT_REFFILE -- Count the number of groups in an image
//
// i: ref	reference file object
//
//------------------------------------------------------------------------
static int count_reffile (reffile *ref)
{
    IRAFPointer ptr;
    int		hi, lo, mid, step;
    char	fname[SZ_FNAME+1];

    Assert(ref != NULL, "count_reffile: BAD ref object");

    // Find a group not in the image 
    hi = 0;
    for (step = 1; step > 0; step *= 2) {
        hi = hi + step;
	sprintf (fname, "%s[%d]", ref->name, hi);

	ptr = c_immap (fname, IRAF_READ_ONLY, 0);
	if (c_iraferr() != 0)
	    break;

	c_imunmap (ptr);
    }

    // Use binary search to find last group in the image 
    lo = hi - step;
    while (hi - lo > 1) {
        mid = (hi + lo) / 2;
	sprintf (fname, "%s[%d]", ref->name, mid);

	ptr = c_immap (fname, IRAF_READ_ONLY, 0);
	if (c_iraferr() != 0) {
	    hi = mid;
	} else {
	    c_imunmap (ptr);
	    lo = mid;
	}
    }

    return (lo);
}

//------------------------------------------------------------------------
// EXT_REFFILE -- Check to see if reference filename has an extension
//                   (e.g. filename[n] )
//
// i: fname	reference file name
//
//------------------------------------------------------------------------
static int reffile_name_has_ext (char *fname)
{
    long	ext;
    char	*lbrak, *rbrak;

    lbrak = strrchr (fname, '[');
    if (lbrak == NULL)
        return (ERR);

    ext = strtol (lbrak+1, &rbrak, 10);
    if (*rbrak != ']')
        return (ERR);

    return (ext);
}

//------------------------------------------------------------------------
// IMAGE_REFFILE -- Routine to read an image header keyword
//
// i: im	image descriptor
// i: name	keyword name
// o: value	keyword value
// i: maxch	max value length
//
// History
// ---------
// 05/20/02  MSwam   Add NAXIS1,2 hack
//
//------------------------------------------------------------------------
static int image_reffile (IRAFPointer im, char *name, char *value, int maxch)
{
    char	buf[SZ_STRING+1], datatype[SZ_STRING+1];
    double	pixval;
    int		pixtype, bitpix, nbytes, axislen;

    cdbs_debug("--image_reffile--");
    switch (strmatch (name, EXTRA_KEYS)) {
    case ORDINARY:
        cdbs_debug("ORDINARY search for %s",name);
	c_imgstr (im, name, value, maxch);

	if (c_iraferr () != 0) {
            cdbs_debug("NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
	break;

    case DATAMIN:
        // SKIP the IRAF special call, and assume the FITS keyword exists...
	// pixval = c_imgmin (im);
	pixval = c_imgetd (im, "DATAMIN");
	if (c_iraferr () != 0) {
            cdbs_debug("DATAMIN keyword NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
        // convert to string
	sprintf (buf, "%f", pixval);
	safecopy (value, buf, maxch);
	break;

    case DATAMAX:
        // SKIP the IRAF special call, and assume the FITS keyword exists...
	// pixval = c_imgmax (im);
	pixval = c_imgetd (im, "DATAMAX");
	if (c_iraferr () != 0) {
            cdbs_debug("DATAMAX keyword NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
        // convert to string
	sprintf (buf, "%f", pixval);
	safecopy (value, buf, maxch);
	break;

    case BITPIX:
        // SKIP the IRAF special call, and assume the FITS keyword exists...
	// pixtype = c_imgtypepix (im);
	// bitpix = IRAFSize[pixtype] * NBITS_BYTE;
	bitpix = c_imgeti (im, "BITPIX");
	if (c_iraferr () != 0) {
            cdbs_debug("BITPIX keyword NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
        // convert to string
	sprintf (buf, "%d", bitpix);
	safecopy (value, buf, maxch);
	break;

    case DATATYPE:
        // let's pretend we can't get this value for now, and see where it
        // fails during testing...
        cdbs_debug("DATATYPE keyword NOT found");
	value[0] = EOS;
	return (ERR);
	//pixtype = c_imgtypepix (im);
	//nbytes = IRAFSize[pixtype];

	//switch (pixtype) {
	//case IRAF_USHORT:
	//    safecopy (datatype, "UNSIGNED", SZ_STRING);
	//    break;
	//case IRAF_SHORT:
	//case IRAF_LONG:
	//case IRAF_INT:
	//    safecopy (datatype, "INTEGER", SZ_STRING);
	//    break;
	//case IRAF_REAL:
	//case IRAF_DOUBLE:
	//    safecopy (datatype, "REAL", SZ_STRING);
	//    break;
	//default:
	//    safecopy (datatype, "REAL", SZ_STRING);
	//    break;
	//}

	//sprintf (buf, "%s*%d", datatype, nbytes);
	//safecopy (value, buf, maxch);
	break;

    case SIMPLE:
	safecopy (value, "F", maxch);
	break;

    case NAXIS1:
        // SKIP the IRAF special call, and assume the FITS keyword exists...
        // axislen = c_imglen(im, 1);
	axislen = c_imgeti (im, "NAXIS1");
	if (c_iraferr () != 0) {
            cdbs_debug("NAXIS1 keyword NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
        if (axislen == 1) {
            //======= HACK =========
            // likely primary header, not the one we want for opr 45589 
            fprintf(stderr,"FYI: NAXIS1 = 1 not being passed to load file.\n");
            value[0] = EOS;
            return (ERR);
        }
        // convert to string
	sprintf (buf, "%d", axislen);
	safecopy (value, buf, maxch);
	break;

    case NAXIS2:
        // SKIP the IRAF special call, and assume the FITS keyword exists...
        // axislen = c_imglen(im, 2);
	axislen = c_imgeti (im, "NAXIS2");
	if (c_iraferr () != 0) {
            cdbs_debug("NAXIS2 keyword NOT found");
	    value[0] = EOS;
	    return (ERR);
	}
        if (axislen == 1) {
            //======= HACK =========
            // likely primary header, not the one we want for opr 45589 
            fprintf(stderr,"FYI: NAXIS2 = 1 not being passed to load file.\n");
            value[0] = EOS;
            return (ERR);
        }
        // convert to string
	sprintf (buf, "%d", axislen);
	safecopy (value, buf, maxch);
	break;

    default:
	cdbs_error (BADLIST, "%s", name);
    }

    return (OK);
}

//-------------------------------------------------------------------------
// RECLOSE_REFFILE -- Close the current image/table/loadfile object
//
// u: ref	reference file object
//
//-------------------------------------------------------------------------
static void reclose_reffile (reffile *ref)
{
    Assert(ref != NULL, "reclose_reffile: BAD ref object");
    if (ref->ptr == NULL)
        return;

    switch (ref->type) {
    case IMAGE:
        c_imunmap ((IRAFPointer) ref->ptr);
	if (c_iraferr () != 0) {
	   cdbs_error (NOCLOSE, "%s", "Failed to close IMAGE file");
        }
	break;
    case TABLE:
	c_tbtclo ((IRAFPointer) ref->ptr);
	if (c_iraferr () != 0) {
	   cdbs_error (NOCLOSE, "%s", "Failed to close TABLE file");
        }
	break;
    case LOADFILE:
	if (ref->smudge == YES)
	    write_loadfile (ref->name, (table *) ref->ptr);

	free_table ((table *) ref->ptr);
	ref->smudge = NO;
	break;
    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    ref->ptr = NULL;
}

//-------------------------------------------------------------------------
// REOPEN_REFFILE -- Open a loadfile or 
//                     a new row (group, extension) in a FITS reference file
//
// u: ref	reference file object
// i: row	row, group, or extension number (zero for primary header)
//
// 09/28/04  52054  MSwam  skipping tbtacc call on Solaris (not working now)
//
//-------------------------------------------------------------------------
static int reopen_reffile (reffile *ref, int row)
{
    Assert(ref != NULL, "reopen_reffile: BAD ref object");

    // if file is open and already at the requested row, do nothing
    if (ref->ptr != NULL && ref->row == row)
        return (OK);

    // if asking for row=1 of a non-FITS file (loadfile)
    // that is currently at the zero-th row, move us to row=1
    //
    if (ref->is_fits == NO && ref->row == 0 && row == 1) {
        ref->row = row;
        return (OK);
    }

    // Don't need to re-open an open FITS table or loadfile to access a new row.
    //   Just set it and return
    //
    if (ref->ptr != NULL && ref->type != IMAGE && 
	ref->row > 0 && row > 0) {
        ref->row = row;
        return (OK);
    }

    // From here we WILL be attempting to open/re-open the file, 
    //   so make sure it is closed
    //
    reclose_reffile (ref);

    // Add the FITS extension to the file name if asking for 
    //   a FITS primary header or an image file is being opened
    //   (NOT for a table, since the [] syntax is appropriated for table row
    //    access instead)
    //
    char fname[SZ_FNAME+1];
    if ((ref->is_fits == YES && row == 0) || ref->type == IMAGE) {
        sprintf (fname, "%s[%d]", ref->name, row);
    } else {
        safecopy (fname, ref->name, SZ_FNAME);
    }

    // Call the appropriate routine to open the file
    cdbs_debug("reopen_reffile: opening %s",fname);
    switch (ref->type) {

    case IMAGE:
        cdbs_debug("reopen_reffile: type IMAGE");
	ref->ptr = (void *) c_immap (fname, (IRAFIOMode) ref->mode, 0);
	if (c_iraferr() != 0) {
	    ref->ptr = NULL;
            fprintf(stderr,"c_immap failed on %s\n",fname);
	    return (ERR);
	}
	break;

    case TABLE:
#ifndef ITS_SPARC_SOLARIS
// THIS IS NOT WORKING ON THE SUNFIRE (?), SO IT WILL BE SKIPPED THERE 
        cdbs_debug("reopen_reffile: type TABLE");
        if (c_tbtacc (fname) == NO) {
            cdbs_debug("c_tbtacc failed on %s\n",fname);
	    return (ERR);
        }
#endif
	ref->ptr = (void *) c_tbtopn (fname, (IRAFIOMode) ref->mode, 0);
	if (c_iraferr() != 0) {
            fprintf(stderr,"c_tbtopn failed on %s\n",fname);
	    ref->ptr = NULL;
	    return (ERR);
	}
	break;

    case LOADFILE:
        cdbs_debug("reopen_reffile: type LOADFILE");
	ref->ptr = (void *) read_loadfile (fname);
	if (ref->ptr == NULL) {
            fprintf(stderr,"read_loadfile failed on %s\n",fname);
	    return (ERR);
	}
	break;

    default:
	cdbs_error (FTYPE, "%d", ref->type);
    }

    // Update row number in object
    ref->row = row;
    return (OK);
}

//------------------------------------------------------------------------
// ROOT_REFFILE -- Extract root (filename minus extension) from filename
//                    e.g. given "fname_typ.fits[0]" returns
//                                fname_typ.fits
//
// o: root	root name
// i: fname	file name, possibly with extension appended (e.g. filename[0])
// i: maxch	max length of root name
//
//------------------------------------------------------------------------
static void root_reffile (char *root, char *fname, int maxch)
{
    char	*lbrak;

    // Copy filename so it can be changed
    safecopy (root, fname, maxch);

    // If filename has an extension look for the start and truncate ext
    if (reffile_name_has_ext (root) != ERR) {
        lbrak = strrchr (root, '[');
	if (lbrak != NULL)
	    lbrak = EOS;
    }
    return;
}
//--------------------------------------------------------------------------
//
// Name: type_reffile
//
// Description: Attempts to determine if a file is in FITS IMAGE, FITS TABLE,
//              or CDBS "loadfile" (ASCII) format.  The file is opened in
//              various ways and tested for certain values that indicate
//              the file type.
//
// Exceptions:
//   std::runtime_error - can be thrown if various unrecoverable errors occur
//
// Returns:
//   ptr to char array (passed in by caller) holding the actual file type
//
// History:
// 03/23/11  MSwam   first version
//--------------------------------------------------------------------------

char *type_reffile (char *fname, char *actual_file_type)
{
    // Try each possible expected file type
    // 
    // case LOADFILE:
    //         look for first 6 chars != 'SIMPLE' (the FITS sentinel)
    FILE *fd;
    if ((fd = fopen (fname, "r")) == NULL) {
        fprintf(stderr,"type_reffile: failed OPEN on %s\n",fname);
        throw std::runtime_error("failed to open file ");
    }
    char line[10];
    if (fgets(line, 10, fd) == NULL) {
        fprintf(stderr,"type_reffile: failed fgets on %s\n",fname);
        fclose (fd);
        throw std::runtime_error("failed to read first line of file ");
    }
    fclose (fd);
    if (strncmp(line,"SIMPLE",6) != 0) {
        cdbs_debug("type_reffile: LOADFILE match (file start != SIMPLE)");
        strcpy(actual_file_type, "load");
        // SUCCESS
        return(actual_file_type);
    }

    // case FITS IMAGE:
    cdbs_debug("type_reffile: trying type IMAGE");
    int status;
    fitsfile *fptr=NULL;
    int cf_iomode = READONLY;
    // the fits open applies to IMAGE or TABLE
    char _ERROR_MSG[256];
    status=0;
    fits_open_file (&fptr, fname, cf_iomode, &status);
    if (status != 0) {
       fits_get_errstatus(status, _ERROR_MSG);
       fprintf(stderr,"fits_open_file status=%d: %s\n",status,_ERROR_MSG);
       cdbs_debug("type_reffile: FITS open failed...");
       throw std::runtime_error("failed to open FITS file ");
    }
    // try to move to the 2nd HDU (1st extension), applies to IMAGE or TABLE
    int hdutype;
    int hdunum = 2;
    /* fits_movabs_hdu = ffmahd */
    status=0;
    fits_movabs_hdu (fptr, hdunum, &hdutype, &status);
    if (status != 0) {
       fprintf(stderr,"type_reffile: failed fits_movabs_hdu to HDU 2 for %s\n",fname);
       // fits_close_file = ffclos
       fits_close_file (fptr, &status);
       throw std::runtime_error("failed to find HDU 2 in FITS file ");
    }
    // try to read the XTENSION keyword value from the 1st extension, 
    //   check for IMAGE or BINTABLE or ASCIITABLE
    //
    char value[SZ_FNAME+1]; /* hopefully large enough for any keyword */
    status = 0;
    /* fits_read_key = ffgky */
    fits_read_key (fptr, TSTRING, "XTENSION", value, NULL, &status);
    if (status != 0) {
       fprintf(stderr,"type_reffile: failed to read KEYWORD=XTENSION from %s\n",
               fname);
       strcpy(actual_file_type,"UNKNOWN");
       fits_close_file (fptr, &status);
       return(actual_file_type);
    }
    if (strcmp(value,"IMAGE")==0) {
       cdbs_debug("type_reffile: IMAGE match");
       strcpy(actual_file_type,"image");
    }
    else {
       // ASSUME TABLE
       cdbs_debug("type_reffile: not LOADFILE or IMAGE, assuming TABLE");
       strcpy(actual_file_type,"table");
    }
    // fits_close_file = ffclos
    fits_close_file (fptr, &status);
    return(actual_file_type);
}
